#!/usr/bin/env python
import os
NE_SRC_PATH = os.path.dirname(os.path.abspath(__file__))

FORTRAN_SRCS = ['dmdsm.NE2001.f', 'density.NE2001.f', 'neLISM.NE2001.f',
                'neclumpN.f', 'nevoidN.f', 'scattering98.f']

DATA_FILES   = ['gal01.inp', 'ne_arms_log_mod.inp', 'ne_gc.inp',
                'nelism.inp', 'neclumpN.NE2001.dat', 'nevoidN.NE2001.dat']

def runcmd(cmd):
    """ Run a command with os.system, printing command to screen"""
    print("\n> " + cmd)
    os.system(cmd)

def list_data_files():
    return [os.path.join(NE_SRC_PATH, df) for df in DATA_FILES]

def cleanup():
    """ """
    print("---- Running cleanup ----")
    runcmd('rm *.o')
    if os.path.exists('sgnFile.pyf'):
        runcmd('rm sgnFile.pyf')


def compile_ne2001(fcomp='gfortran', ar_flags='rc'):
    orig_cwd = os.getcwd()
    try:
        os.chdir(NE_SRC_PATH)

        print("\n##########################")
        print("#### Compiling NE2001 ####")
        print("##########################\n")
        print("Working directory: " + NE_SRC_PATH)
        print("Fortran compiler: " + fcomp)

        AR_CMD = 'ar {ar_flags} libNE2001.a '.format(ar_flags=ar_flags)
        RANLIB_CMD = 'ranlib libNE2001.a'

        for fsrc in FORTRAN_SRCS:
            fobj = fsrc.replace('.f', '.o')
            runcmd("{fcomp} -O -std=gnu -c -o {fobj} {fsrc}".format(fcomp=fcomp, fobj=fobj, fsrc=fsrc))
            AR_CMD += '{fobj} '.format(fobj=fobj)

        runcmd(AR_CMD)
        runcmd(RANLIB_CMD)

        print("\n---- Generating F2PY shared object for DMDSM ----")
        runcmd('f2py -m dmdsm -h sgnFile.pyf dmdsm.NE2001.f --overwrite-signature')
        runcmd('f2py -c sgnFile.pyf dmdsm.NE2001.f -L./ -lNE2001 -m dmdsm')
        runcmd('rm sgnFile.pyf')

        print("\n---- Generating F2PY shared object for density ----")
        runcmd('f2py -m density -h sgnFile.pyf density.NE2001.f --overwrite-signature')
        runcmd('f2py -c sgnFile.pyf density.NE2001.f -L./ -lNE2001 -m density')
        runcmd('rm sgnFile.pyf')
    except:
        raise
    finally:
        os.chdir(orig_cwd)

if __name__ == "__main__":
    try:
        compile_ne2001()
        print("Complete.")
    except:
        raise
    finally:
        cleanup()






import numpy as np
import os
from functools import wraps
from astropy import units as u
from . import dmdsm     # f2py FORTRAN object
from . import density   # f2py FORTRAN object

DATA_PATH = os.path.dirname(os.path.abspath(__file__))

"""
    dmdsm f2py object
    --------------------
    Call signature: dmdsm.dmdsm(*args, **kwargs)
    Type:           fortran
    String form:    <fortran object>
    Docstring:
    limit,sm,smtau,smtheta,smiso = dmdsm(l,b,ndir,dmpsr,dist)

    Wrapper for ``dmdsm``.

    Parameters
    ----------
    l : input float
    b : input float
    ndir : input int
    dmpsr : in/output rank-0 array(float,'f')
    dist : in/output rank-0 array(float,'f')

    Returns
    -------
    limit : string(len=1)
    sm : float  (scattering measure, uniform weighting) (kpc/m^{20/3})
    smtau : float  (scattering measure, weighting for pulse broadening)
    smtheta : float  (scattering measure, weighting for angular broadening of galactic sources)
    smiso : float (scattering measure appropriate for calculating the isoplanatic angle at the source's location)


    density f2py object
    --------------------
    Call signature: density.density_2001(*args, **kwargs)
    Type:           fortran
    String form:    <fortran object>
    Docstring:
    ne1,ne2,nea,negc,nelism,necn,nevn,f1,f2,fa,fgc,flism,fcn,fvn,whicharm,wlism,wldr,
    wlhb,wlsb,wloopi,hitclump,hitvoid,wvoid = density_2001(x,y,z)

    Wrapper for ``density_2001``.

    Parameters
    ----------
    x : input float
    y : input float
    z : input float

    Returns
    -------
    ne1 : float
    ne2 : float
    nea : float
    negc : float
    nelism : float
    necn : float
    nevn : float
    f1 : float
    f2 : float
    fa : float
    fgc : float
    flism : float
    fcn : float
    fvn : float
    whicharm : int
    wlism : int
    wldr : int
    wlhb : int
    wlsb : int
    wloopi : int
    hitclump : int
    hitvoid : int
    wvoid : int

"""


def run_from_pkgdir(f):
    """ chdir() into package directory when running

    Fortran code doesn't know the relative path to its data files.
    This wraps the function call, changing into the right directory
    first, calling it, then changing back to original directory.
    """
    @wraps(f)
    def wrapped(*args, **kwargs):
        cwdpath = os.getcwd()
        try:
            os.chdir(DATA_PATH)
            r = f(*args, **kwargs)
            return r
        except:   # pragma: no cover
            raise
        finally:
            os.chdir(cwdpath)
    return wrapped


def TAUISS(d, sm, nu):
    """ Convert a scattering measure (SM) to scattering timescale at given frequency.

    Direct port from FORTRAN code scattering98.f

    Args:
        d (float): Distance in kpc
        sm (float): Scattering measure (kpc m^{-20/3})
        nu (float): Radio frequency in GHz

    Returns:
        tauiss (float): pulse broadening time (ms)

    Fortran equiv:
          REAL FUNCTION TAUISS(d, sm, nu)
    c
    c calculates the pulse broadening time in ms
    c from distance, scattering measure, and radio frequency
    c
    c input:      d = pulsar distance       (kpc)
    c            sm = scattering measure    (kpc m^{-20/3})
    c            nu = radio frequency       (GHz)
    c output: tausis = pulse broadening time (ms)
    c
          implicit none
          real d, sm,  nu
          tauiss = 1000. * (sm / 292.)**1.2 * d * nu**(-4.4)
          end
    """
    tauiss = 1. * (sm / 292.)**1.2 * d * nu**(-4.4)
    return tauiss


@run_from_pkgdir
def dm_to_dist(l, b, dm):
    """ Convert DM to distance and compute scattering timescale
    
    Args:
        l (float): galactic latitude in degrees
        b (float): galactic longitude in degrees
        dm (float or np.array): Dispersion measure
    """
    dm    = np.array(dm, dtype='float32')
    dist  = np.zeros_like(dm)
    l_rad = np.deg2rad(l)
    b_rad = np.deg2rad(b)
    ndir = 1
    if np.isclose(dm, 0):  # Catch infinite timeout bug
        return 0.0 * u.pc, 0.0 * u.s
    else:
        limit,sm,smtau,smtheta,smiso = dmdsm.dmdsm(l_rad,b_rad,ndir,dm,dist)

    tau_sc =TAUISS(float(dist), smtau, nu=1.0)
    return (float(dist) * u.kpc).to('pc'), tau_sc * u.s


@run_from_pkgdir
def dist_to_dm(l, b, dist):
    """ Convert distance to DM and compute scattering timescale
    
    Args:
        l (float): galactic latitude in degrees
        b (float): galactic longitude in degrees
        dm (float or np.array): Dispersion measure
    """
    dist  = np.array(dist, dtype='float32')
    dm    = np.zeros_like(dist)
    l_rad = np.deg2rad(l)
    b_rad = np.deg2rad(b)
    ndir = -1

    if np.isclose(dist, 0):       # Catch infinite timeout bug
        return 0.0 * u.pc / u.cm**3, 0.0 * u.s
    else:
        limit,sm,smtau,smtheta,smiso = dmdsm.dmdsm(l_rad,b_rad,ndir,dm,dist)

    tau_sc = TAUISS(float(dist), smtau, nu=1.0)
    return float(dm) * u.pc / u.cm**3, tau_sc * u.s


@run_from_pkgdir
def calculate_electron_density_xyz(x, y, z):
    """ Compute electron density at Galactocentric X, Y, Z coordinates 

    x,y,z are Galactocentric Cartesian coordinates, measured in kiloparsecs,
    with the axes parallel to (l, b) = (90, 0), (180, 0), and (0, 90) degrees

    Args:
        x, y, z (float): Galactocentric coordinates in kpc (NOT pc!)
    """
    ne_out = density.density_2001(x, y, z)
    return np.sum(ne_out[:7]) / u.cm**3







"""
# yt2020.py - Python implementation of Yamasaki & Totani DM Halo model

References:
    [1] Yamasaki S, Totani T,
    'The Galactic Halo Contribution to the Dispersion Measure of Extragalactic Fast Radio Bursts'
    https://ui.adsabs.harvard.edu/abs/2019arXiv190900849Y/abstract
    2019, arXiv e-prints, arXiv:1909.00849

Notes:
    Adapted from S. Yamasaki's DM_halo_yt2020_numerical.py command-line python code
"""
import numpy as np
from scipy import integrate
from numpy import cos, sin, sqrt, fabs
from astropy import units as u

### numerical constants ###

U_G    = 6.67                           # G = 6.67259e-8 cm^3/g/s^2
U_Msun = 1.99                           # 1 M_sun = 1.99e+33 g
U_pc   = 3.09                           # 1 pc = 3.086e+18 cm
U_mp   = 1.67                           # m_p = 1.6726231e-24 g
U_eV   = 1.60                           # 1 eV = 1.602177e-12 erg
c_scaling = U_Msun/U_pc**3*pow(10,-18)  # scaling factor [10^{12} M_sun/kpc^3] -> [g/cm^3]
m_p    = 1.6726231*pow(10,-24)          # proton mass [g]

### key parameters ###

T      = 0.3                    # halo temperature [keV]
M_b    = 0.036                  # total halo gas mass [10^{12} Z^{-1} M_sun] (such that M_b = 0.12 for Z = Z_halo)
Z_halo = 0.3                    # halo metallicity [Z_sun]
xi_H   = 28.0/34.0              # abundance ratio of hydrogen to electron
mu_e   = 1.18                   # mean molecular mass per electron
mu     = 0.62                   # mean particle mass
D_sun  = 8.5                    # Galactocentoric distance of the Sun [kpc]
c_NFW  = 12                     # NFW concentration
r_vir  = 260                    # virial radius of MW [kpc]
M_vir  = 1.0                    # virial mass of MW [10^{12} M_sun]
r_s    = r_vir/c_NFW            # NFW scale radius [kpc]
f_NFW  = np.log(1.0+c_NFW)-c_NFW/(1.0+c_NFW) # f(x)=np.log(1.0+x)-x/(1.0+x) and x = c_NFW
rho_s  = M_vir/4.0/np.pi/r_s**(3.0)/f_NFW    # NFW scale density [10^{12} M_sun/kpc^3]

### calculate n_0^{sphe} (eq.[3]) by normalization (eq.[4]) ###

r_array = np.linspace(r_s/100,r_vir,100000)                         # r = 0-r_vir [kpc]
Upsilon = 4*np.pi*10*U_G*U_Msun/U_pc*U_mp/U_eV*r_s**2*rho_s*mu/T    # Upsilon appearing in right-hand side of eq.(3)

# integrated function excluding n_0^{sphe} (left-hand side of eq.[4])
y_sphe  = 4.0*np.pi*r_array**2*mu_e*m_p/c_scaling*np.exp(-Upsilon*(1-np.log(1+r_array/r_s)*r_s/r_array))

I_sphe   = integrate.simps(y_sphe,r_array)      # integration
n_0_sphe = M_b/Z_halo/I_sphe                    # central electron number density for spherical comp. [cm^{-3}]
#print('Upsilon = %f' % Upsilon)
#print('n_0_sphe = %f [cm^{-3}]' % n_0_sphe)

### key functions ###


def s_max(l, b): # maximum integration limit corresponsing to r = r_vir as function of sky coordinates [kpc]
    """ Compute integration limit s_max for given sky coordinates

    Args:
        l (float): Galactic longitude, in radians (-pi to +pi)
        b (float): Galactic latitude, in radians  (-pi/2 to pi/2)

    Returns:
        s_max (float), maximum integration limit corresponsing to r = r_vir
    """
    return D_sun*cos(b)*cos(l)+sqrt(D_sun**2*(cos(b)**2*cos(l)**2-1)+(1.0*r_vir)**2)


def ne_sphe(l, b, s):
    """ Compute electron density for spherical component for (l, b) at distance s
    Args:
        l (float): Galactic longitude, in radians (-pi to +pi)
        b (float): Galactic latitude, in radians  (-pi/2 to pi/2)
        s (float): Distance (kpc)

    Returns:
        ne (float): electron density in cm^{-3}
    """
    R = sqrt(D_sun**2+(s*cos(b))**2-2*D_sun*s*cos(b)*cos(l))
    z = s*sin(b)
    r = sqrt(R**2+z**2)
    Upsilon = 4*np.pi*10*U_G*U_Msun/U_pc*U_mp/U_eV*r_s**2*rho_s*mu/T
    return n_0_sphe*np.exp(-Upsilon*(1-np.log(1+r/r_s)*r_s/r)) # [cm^{-3}]


def ne_disk(l, b, s):
    """ Compute electron density for spherical component for (l, b) at distance s
    Args:
        l (float): Galactic longitude, in radians (-pi to +pi)
        b (float): Galactic latitude, in radians  (-pi/2 to pi/2)
        s (float): Distance (kpc)

    Returns:
        ne (float): electron density in cm^{-3}
    """
    R = sqrt(D_sun**2+(s*cos(b))**2-2*D_sun*s*cos(b)*cos(l))
    z = fabs(s*sin(b))
    n_0_disk = 7.4*pow(10,-3) # [Z^{-1} cm^{-3}]
    z_0 = 2.4 # [kpc]
    R_0 = 4.9 # [kpc]
    return n_0_disk/np.exp(R/R_0+z/z_0)/Z_halo # [cm^{-3}]


### vectorize density profile functions ###
vfunc_sphe = np.vectorize(ne_sphe)
vfunc_disk = np.vectorize(ne_disk)

def calculate_halo_dm(l, b, component='both'):
    """ Compute halo DM

    Args:
        l (float): Galactic longitude, in degrees (-180 to +180)
        b (float): Galactic latitude, in degrees  (-90 to 90)
        component (str): Compute 'spherical' component of halo,
                         'disk' component, or 'both' components.

    Returns:
        DM (float): Dispersion measure in [pc/cm^3]
    """

    ### input coordinates (l,b) in degrees ###
    l_FRB = np.deg2rad(l)
    b_FRB = np.deg2rad(b)

    ### calculate DM_halo ###
    N = int(10000)
    xx = np.linspace(0.01, s_max(l_FRB, b_FRB), N)
    y_DM_sphe = vfunc_sphe(l_FRB, b_FRB, xx)
    y_DM_disk = vfunc_disk(l_FRB, b_FRB, xx)

    if component == 'spherical' or component == 'both':
        DM_sphe = 1000 * integrate.simps(y_DM_sphe, xx) * u.pc / u.cm**3 # [pc/cm^3]
    if component == 'disk' or component == 'both':
        DM_disk = 1000 * integrate.simps(y_DM_disk, xx) * u.pc / u.cm**3 # [pc/cm^3]

    if component == 'both':
        return DM_sphe + DM_disk
    elif component == 'spherical':
        return DM_sphe
    elif component == 'disk':
        return DM_disk
    else:
        raise RuntimeError("Component must be 'both', 'spherical', or 'disk' ")


def calculate_halo_dm_analytic(l, b):
    """ Calculate the DM contribution of the Galactic halo.

    Uses an analytical formula for speed. Useful for all-sky mapping.

    Args:
        l (float): Galactic longitude, in degrees (-180 to +180)
        b (float): Galactic latitude, in degrees  (-90 to 90)
    """
    YT19_coeffs = np.array([[250.12, -871.06, 1877.5, -2553.0, 2181.3, -1127.5, 321.72, -38.905],
                            [-154.82, 783.43, -1593.9, 1727.6, -1046.5, 332.09, -42.815, 0],
                            [-116.72, -76.815, 428.49, -419.00, 174.60, -27.610, 0, 0],
                            [216.67, -193.30, 12.234, 32.145, -8.3602, 0, 0, 0],
                            [-129.95, 103.80, -22.800, 0.44171, 0, 0, 0, 0],
                            [39.652, -21.398, 2.7694, 0, 0, 0, 0, 0],
                            [-6.1926, 1.6162, 0, 0, 0, 0, 0, 0],
                            [0.39346, 0, 0, 0, 0, 0, 0, 0]])
    l_abs = np.abs(np.deg2rad(l))
    b_abs = np.abs(np.deg2rad(b))
    DM_halo = 0
    for ii in range(8):
        for jj in range(8):
            DM_halo += YT19_coeffs[ii, jj] * l_abs ** ii * b_abs ** jj
    return DM_halo * u.pc / u.cm**3


#!/usr/bin/env python
"""
# PyGEDM.py -- Galactic Electron Density Models (NE2001 + YMW16) in python
"""

from . import ymw16_wrapper
from . import ne2001_wrapper
from . import yt2020
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord, Galactocentric
from astropy.units import Quantity, Unit
import numpy as np
from pkg_resources import resource_filename, get_distribution, DistributionNotFound

try:
    import healpy as hp
    HAS_HEALPIX = True
except ImportError:  # pragma: no cover
    HAS_HEALPIX = False

try:
    __version__ = get_distribution('pygedm').version
except DistributionNotFound: # pragma: no cover1
    __version__ = '0.0.1 - manual install'


def _gl_gb_convert(gl, gb, unit='rad'):
    """ Convert gl, gb astropy.Angle to floats/arrays with correct units

    If not an astropy quantity, returns value unchanged.
    """
    if isinstance(gl, Angle) or isinstance(gl, Quantity):
        gl = gl.to(unit).value
    if isinstance(gb, Angle) or isinstance(gb, Quantity):
        gb = gb.to(unit).value
    return gl, gb


def _unit_convert(q, unit):
    """ Convert astropy.unit into float/array with given unit

    If not an astropy quantity, returns value unchanged.
    """
    if isinstance(q, Quantity):
        q = q.to(unit).value
    return q

def dm_to_dist(gl, gb, dm, dm_host=0, mode='gal', method='ymw16'):
    """ Convert a DM to a distance

    Args:
        gl: galactic longitude (float in deg or astropy.Angle)
        gb: galactic latitude (float in deg or astropy.Angle)
        dm: dispersion measure (float or astropy.Quantity, , pc cm^-3)
        mode: Gal, MC, or IGM

    Returns:
        (dist, tau_sc) tuple
        dist: distance (pc, astropy.Quantity)
        tau_sc: scattering time scale at 1 GHz (s, astropy.Quantity)
    """
    gl, gb = _gl_gb_convert(gl, gb, 'deg')
    dm = _unit_convert(dm, 'pc cm^(-3)')

    if method.lower() == 'ymw16':
        return ymw16_wrapper.dm_to_dist(gl, gb, dm, dm_host=0, mode=mode)
    elif method.lower() == 'ne2001':
        if mode != 'gal':
            raise RuntimeError('NE2001 only supports Galactic (gal) mode.')
        return ne2001_wrapper.dm_to_dist(gl, gb, dm - dm_host)
    else:
        raise RuntimeError("Only ymw16 and ne2001 models supported.")


def dist_to_dm(gl, gb, dist, mode='gal', method='ymw16'):
    """ Convert a distance to a DM

    Args:
        gl: galactic longitude (deg)
        gb: galactic latitude (deg)
        dist: distance to source (pc) or if in mode IGM (Mpc)

    Returns:
        (dm, tau_sc) tuple
        dm: dispersion measure (float or astropy.Quantity, , pc cm^-3)
        tau_sc: scattering time scale at 1 GHz (s, astropy.Quantity)
    """
    gl, gb = _gl_gb_convert(gl, gb, 'deg')

    if mode == 'igm':
        dist = _unit_convert(dist, 'Mpc')
    else:
        dist = _unit_convert(dist, 'pc')

    if method.lower() == 'ymw16':
        return ymw16_wrapper.dist_to_dm(gl, gb, dist, mode=mode)
    elif method.lower() == 'ne2001':
        if mode != 'gal':
            raise RuntimeError('NE2001 only supports Galactic (gal) mode.')
        dist_kpc = dist / 1000.0
        return ne2001_wrapper.dist_to_dm(gl, gb, dist_kpc)
    else:
        raise RuntimeError("Only ymw16 and ne2001 models supported.")


def calculate_electron_density_xyz(x, y, z, method='ymw16'):
    """ Calculate electron density at a point with galactocentric coords (x, y, z)

    Args:
        (x, y, z): galactocentric coordinates in pc

    Returns:
        N_e: electron density in cm^-3
    """
    x = _unit_convert(x, 'pc')
    y = _unit_convert(y, 'pc')
    z = _unit_convert(z, 'pc')

    if method.lower() == 'ymw16':
        return ymw16_wrapper.calculate_electron_density_xyz(x, y, z)
    elif method.lower() == 'ne2001':
        return ne2001_wrapper.calculate_electron_density_xyz(x/1e3, y/1e3, z/1e3)
    else:
        raise RuntimeError("Only ymw16 and ne2001 models supported.")


def calculate_electron_density_lbr(gl, gb, dist, method='ymw16'):
    """ Calculate electron density at a point with Galactic coords (ga, gl)
        at a given distance in pc

    Args:
        (gl, gb, dist): Galactic lat/long in deg, dist in pc

    Returns:
        N_e: electron density in cm^-3
    """
    gl, gb = _gl_gb_convert(gl, gb, 'deg')
    dist = _unit_convert(dist, 'pc')

    if method.lower() == 'ymw16':
        return ymw16_wrapper.calculate_electron_density_lbr(gl, gb, dist)
    elif method.lower() == 'ne2001':
        x, y, z = convert_lbr_to_xyz(gl, gb, dist, method='ne2001')
        return ne2001_wrapper.calculate_electron_density_xyz(x.to('kpc').value, y.to('kpc').value, z.to('kpc').value)
    else:
        raise RuntimeError("Only ymw16 and ne2001 models supported.")


def convert_lbr_to_xyz(gl, gb, dist, method='ymw16'):
    """ Convert Galactic (l,b,r) coords to Galactocentric (x,y,z) coords

    Args:
        (gl, gb, dist): Galactic lat/long in deg, dist in pc.
                        use of astropy Angle/Quantities also OK
        method (str): one of 'ymw16', 'ne2001', or 'astropy'

    Notes:
        For transform, the Sun is located at (x=0, y=R_sun, z=z_sun)
        YMW16  uses R_sun of 8300 pc and z_sun of 6.0 pc
        NE2001 uses R_sun of 8500 pc and z_sun of 0.0 pc
        Both of these do a basic spherical to cartesian conversion.

        astropy does a much more complicated conversion, see
        https://astropy.readthedocs.io/en/latest/coordinates/galactocentric.html
        This is the 'proper' coordinate system, but note that it is NOT COMPATIBLE
        WITH NE2001 OR YMW16! (!SEE EXAMPLE OUTPUT BELOW!)

    Example output:
        pygedm.convert_lbr_to_xyz(0, 0, 0, method='ymw16')
        (<Quantity 0. pc>, <Quantity 8300. pc>, <Quantity 6. pc>)

        pygedm.convert_lbr_to_xyz(0, 0, 0, method='ne2001')
        (<Quantity 0. pc>, <Quantity 8500. pc>, <Quantity 0. pc>)

        pygedm.convert_lbr_to_xyz(0, 0, 0, method='astropy')
        (<Quantity -8499.95711754 pc>, <Quantity 0. pc>, <Quantity 27. pc>)
    """

    gl, gb = _gl_gb_convert(gl, gb, 'deg')
    gl = np.deg2rad(gl)
    gb = np.deg2rad(gb)

    dist = _unit_convert(dist, 'pc') # Unit conversion and strip quantity
    dist = dist * u.pc               # Make sure distance is in units of pc
    if method == 'astropy':
        lbr = SkyCoord(gl, gb, distance=dist, frame='galactic', unit=('degree', 'degree', 'pc'))
        xyz = lbr.transform_to(Galactocentric(galcen_distance=8.5 * u.kpc))
        return xyz.x, xyz.y, xyz.z
    elif method.lower() == 'ymw16':
        Rsun = 8300 * u.pc
        Zsun = 6.0 * u.pc
        x = dist * np.sin(gl) * np.cos(gb)
        y = Rsun - dist * np.cos(gl) * np.cos(gb)
        z = Zsun + dist * np.sin(gb)
        return x, y, z
    elif method.lower() == 'ne2001':
        Rsun = 8500 * u.pc
        x = dist * np.sin(gl) * np.cos(gb)
        y = Rsun - dist * np.cos(gl) * np.cos(gb)
        z = dist * np.sin(gb)
        return x, y, z
    else:
        raise RuntimeError("method must be astropy, ne2001, or ymw16")


def generate_healpix_dm_map(dist=1, nside=64, method='ymw16'):
    """ Generate an all-sky healpix map for a given distance.

    Args:
        dist (float or Quantity): Distance to integrate EDM out to. 30 kpc will go to edge
        nside (int): The NSIDE parameter for the healpix map (power of 2, larger => higher resolution)
        method (str): one of 'ymw16', 'ne2001', 'yt2020' or 'yt2020_analytic'

    Notes:
        This method takes a fair amount of time to run -- tens of seconds for NSIDE=32.
        YT2020 method is even slower, consider using yt2020_analytic

    Returns:
         healpix map as a numpy array (1D), which can be viewed using the healpy.mollview() method
    """
    if HAS_HEALPIX:
        pix_id = np.arange(hp.nside2npix(nside))
        gl, gb = hp.pix2ang(nside, pix_id, lonlat=True)
        d = np.zeros_like(pix_id, dtype='float32')
        if 'yt2020' not in method:
            for ii in pix_id:
                dm, tau_sc = dist_to_dm(gl[ii], gb[ii], dist, method=method)
                d[ii] = dm.value

        elif 'yt2020' in method:
            # gl is in (0, 360), wrap to (-180, 180) as reqd for yt2020 function
            # Then call yt2020.compute_halo_dm function directly
            gl_u = np.copy(gl)
            gl_u[gl >= 180] -= 360
            if method == 'yt2020':
                for ii in pix_id:
                    dm_halo = yt2020.calculate_halo_dm(gl_u[ii], gb[ii])
                    d[ii] = dm_halo.value
            else:
                for ii in pix_id:
                    dm_halo = yt2020.calculate_halo_dm_analytic(gl_u[ii], gb[ii])
                    d[ii] = dm_halo.value
        return d
    else:
        raise RuntimeError("Healpy installation not found.")


def calculate_halo_dm(gl, gb, method='yt2020', component='both'):
    """ Compute halo DM
    Args:
        (gl, gb): Galactic lat/long in degrees
                        use of astropy Angle/Quantities also OK
        method (str): one of 'yt2020' (only YT2020 supported currently)
        component (str): Compute 'spherical' component of halo,
                         'disk' component, or 'both' components.
    Returns:
        DM (float): Dispersion measure in [pc/cm^3]

    """
    gl, gb = _gl_gb_convert(gl, gb, 'deg')
    gl = Angle(gl, unit='deg').wrap_at('180d').value  ## Ensure in correct range

    if method == 'yt2020':
        return yt2020.calculate_halo_dm(gl, gb, component)
    elif method == 'yt2020_analytic':
        if component != 'both':
            raise RuntimeError("Analytic formula requires component='both'")
        return yt2020.calculate_halo_dm_analytic(gl, gb)
    else:
        raise RuntimeError("Only YT2020 currently supported.")


# To increment version
# Check you have ~/.pypirc filled in
# git tag x.y.z
# git push --tags
# python setup.py sdist bdist_wheel
# twine upload dist/*

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import sys
import os
import glob
import setuptools

__version__ = '3.1.1'
__here__ = os.path.abspath(os.path.dirname(__file__))

if sys.version_info.major == 3:
    astro = "astropy"
else:
    astro = "astropy<3.0"

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        'ymw16',
        sources=[
            'ymw16_src/main.cpp',
            'ymw16_src/dora.cpp',
            'ymw16_src/fermibubble.cpp',
            'ymw16_src/frb_d.cpp',
            'ymw16_src/galcen.cpp',
            'ymw16_src/gum.cpp',
            'ymw16_src/lmc.cpp',
            'ymw16_src/localbubble.cpp',
            'ymw16_src/ne_crd.cpp',
            'ymw16_src/nps.cpp',
            'ymw16_src/smc.cpp',
            'ymw16_src/spiral.cpp',
            'ymw16_src/thick.cpp',
            'ymw16_src/thin.cpp',
            'ymw16_src/ymw16par.cpp',
            'ymw16_src/dmdtau2.cpp',
        ],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(__here__, 'ymw16_src'),
        ],
        extra_link_args=['-lm'],
        language='c++'
    ),
]

ymw16_data_files  = ['spiral.txt', 'ymw16par.txt']
ne2001_data_files = ['gal01.inp', 'ne_arms_log_mod.inp', 'ne_gc.inp',
                     'nelism.inp', 'neclumpN.NE2001.dat', 'nevoidN.NE2001.dat']
data_files = ymw16_data_files + ne2001_data_files

# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


with open("README.md", "r") as fh:
    long_description = fh.read()

# Compile NE2001
NE_SRC_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'ne2001_src')
FORTRAN_SRCS = ['dmdsm.NE2001.f', 'density.NE2001.f', 'neLISM.NE2001.f',
                'neclumpN.f', 'nevoidN.f', 'scattering98.f']
DATA_FILES   = ['gal01.inp', 'ne_arms_log_mod.inp', 'ne_gc.inp',
                'nelism.inp', 'neclumpN.NE2001.dat', 'nevoidN.NE2001.dat']


def runcmd(cmd):
    """ Run a command with os.system, printing command to screen"""
    print("\n> " + cmd)
    os.system(cmd)


def list_data_files():
    return [os.path.join(NE_SRC_PATH, df) for df in DATA_FILES]


def cleanup():
    """ """
    print("---- Running cleanup ----")
    runcmd('rm *.o')
    if os.path.exists('sgnFile.pyf'):
        runcmd('rm sgnFile.pyf')


def compile_ne2001(fcomp='gfortran', ar_flags='rc'):
    orig_cwd = os.getcwd()
    try:
        os.chdir(NE_SRC_PATH)

        print("\n##########################")
        print("#### Compiling NE2001 ####")
        print("##########################\n")
        print("Working directory: " + NE_SRC_PATH)
        print("Fortran compiler: " + fcomp)

        AR_CMD = 'ar {ar_flags} libNE2001.a '.format(ar_flags=ar_flags)
        RANLIB_CMD = 'ranlib libNE2001.a'

        for fsrc in FORTRAN_SRCS:
            fobj = fsrc.replace('.f', '.o')
            runcmd("{fcomp} -O -fPIC -std=gnu -c -o {fobj} {fsrc}".format(fcomp=fcomp, fobj=fobj, fsrc=fsrc))
            AR_CMD += '{fobj} '.format(fobj=fobj)

        runcmd(AR_CMD)
        runcmd(RANLIB_CMD)

        print("\n---- Generating F2PY shared object for DMDSM ----")
        runcmd('f2py -m dmdsm -h sgnFile.pyf dmdsm.NE2001.f --overwrite-signature')
        runcmd('f2py -c sgnFile.pyf dmdsm.NE2001.f -L./ -lNE2001 -m dmdsm')
        runcmd('rm sgnFile.pyf')

        print("\n---- Generating F2PY shared object for density ----")
        runcmd('f2py -m density -h sgnFile.pyf density.NE2001.f --overwrite-signature')
        runcmd('f2py -c sgnFile.pyf density.NE2001.f -L./ -lNE2001 -m density')
        runcmd('rm sgnFile.pyf')
    except:
        raise
    finally:
        os.chdir(orig_cwd)

####
# COMPILE NE2001 FORTRAN
####


do_fortran_compile = True

if do_fortran_compile:
    compile_ne2001()

ne2001_shared_objs = list(glob.glob(NE_SRC_PATH + '/*.so'))
for sobj in ne2001_shared_objs:
    print("Copying {sobj} to pygedm/".format(sobj=os.path.basename(sobj)))
    os.system('cp {sobj} pygedm/'.format(sobj=sobj))
    data_files.append(os.path.basename(sobj))

setup(
    name='pygedm',
    version=__version__,
    author='D. C. Price',
    author_email='dancpr@berkeley.edu',
    description='Python/C++ version of YMW16 electron density model',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/telegraphic/pygedm',
    download_url='https://github.com/telegraphic/pygedm/archive/%s.tar.gz' % __version__,
    python_requires='>=2.7',
    install_requires=['pybind11>=2.2', astro],
    tests_require= ['pytest<3.7', astro, 'numpy', 'healpy'],
    setup_requires= ['pytest-runner', 'pytest-cov', 'pybind11>=2.2'],
    ext_modules=ext_modules,
    packages=['pygedm'],
    package_data={'pygedm': data_files},
    include_package_data=True,
    zip_safe=False,
    cmdclass={'build_ext': BuildExt},
)



#!/usr/bin/env python
"""
# PyYMW16.py -- python/C++ port of YMW16 C code
"""

import ymw16
from astropy import units as u
from astropy.coordinates import Angle
from astropy.units import Quantity, Unit
import os
from pkg_resources import resource_filename, get_distribution, DistributionNotFound

DATAPATH = os.path.dirname(resource_filename("pygedm", "spiral.txt"))


MODE_IDS = {'gal': 1,
            'mc': 0,
            'igm': -1}


def dm_to_dist(gl, gb, dm, dm_host=0, mode='gal'):
    """ Convert a DM to a distance

    Args:
        gl: galactic longitude (float in deg or astropy.Angle)
        gb: galactic latitude (float in deg or astropy.Angle)
        dm: dispersion measure (float or astropy.Quantity, , pc cm^-3)
        mode: Gal, MC, or IGM

    Returns:
        (dist, tau_sc) tuple
        dist: distance (pc, astropy.Quantity)
        tau_sc: scattering time scale (s, astropy.Quantity)
    """

    mode_id = MODE_IDS.get(mode.lower().strip())
    ndir, vbs, txt = 1, 0, ''

    r = ymw16.dmdtau(gl, gb, dm, dm_host, ndir, mode_id, vbs, DATAPATH, txt)
    if mode == 'igm':
        r['dist'] *= u.Mpc
        r['tau_sc'] = r['tau_FRB']
    else:
        r['dist'] *= u.pc
    dist = r['dist']
    tau_sc = r['tau_sc'] * u.s
    return (dist, tau_sc)


def dist_to_dm(gl, gb, dist, mode='gal'):
    """ Convert a distance to a DM

    Args:
        gl: galactic longitude (deg)
        gb: galactic latitude (deg)
        dist: distance to source (pc) or if in mode IGM (Mpc)
    """
    mode_id = MODE_IDS.get(mode.lower().strip())
    ndir, dm_host, vbs, txt = 2, 0, 0, ''

    r = ymw16.dmdtau(gl, gb, dist, dm_host, ndir, mode_id, vbs, DATAPATH, txt)
    if mode == 'igm':
        r['DM'] = r['DM_IGM']
        r['tau_sc'] = r['tau_FRB']
    dm = r['DM'] * u.pc / u.cm**3
    tau_sc = r['tau_sc'] * u.s
    return dm, tau_sc


def calculate_electron_density_xyz(x, y, z):
    """ Calculate electron density at a point with galactocentric coords (x, y, z)

    Args:
        (x, y, z): galactocentric coordinates in pc

    Returns:
        N_e: electron density in cm^-3
    """
    ndir, vbs, txt = 1, 0, ''
    ne = ymw16.ne_crd(x, y, z, 0, 0, 0, ndir, vbs, DATAPATH, txt)
    return ne / u.cm**3


def calculate_electron_density_lbr(gl, gb, dist):
    """ Calculate electron density at a point with Galactic coords (ga, gl)
        at a given distance in pc

    Args:
        (gl, gb, dist): Galactic lat/long in deg, dist in pc

    Returns:
        N_e: electron density in cm^-3
    """
    ndir, vbs, txt = 2, 0, ''
    ne = ymw16.ne_crd(0, 0, 0, gl, gb, dist, ndir, vbs, DATAPATH, txt)
    return ne  / u.cm**3
